/*
 * JBoss, Home of Professional Open Source
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2008
 * @author JBoss Inc.
 */
package org.jboss.jbossts.rts.example;

import org.jboss.jbossts.rts.provider.TMUnavailableException;
import org.jboss.jbossts.rts.provider.HttpResponseException;
import org.jboss.jbossts.rts.provider.TransactionStatusException;
import org.jboss.jbossts.rts.util.TxUtil;
import org.jboss.jbossts.rts.service.*;
import org.jboss.resteasy.util.HttpResponseCodes;
import org.apache.log4j.Logger;

import javax.ws.rs.*;
import javax.ws.rs.core.Response;
import java.util.*;

/**
 * Participants are responsible for doing work on real resources and for obeying a 2PC protocol.
 * The interface is defined by TransactionalParticipant.
 * Participant URLs are unique for each transaction but is unconstrained. To satisfy this restriction
 * we choose encode the transaction id into the url:
 *         participant{pId}/tx/{tid}
 * where pId is a participant id and tid is the transaction id (actually only tid would do).
 */
@Path("/participant")
public class Participant //implements TransactionalParticipant
{
    protected final static Logger log = Logger.getLogger(Participant.class);

    /*
     * Define methods for spec conformance
     */

    /**
     * Performing a GET on the participant URL will return the current status
     * of the participant, or 404 if the participant is no longer present.
     */
    @GET
    @Path("{pId}/tx/{tid}/{status}")
    public Response getStatus(@PathParam("pId")String pid,
                              @PathParam("tid") @DefaultValue("")String tid,
                              @PathParam("status") String status)
    {
        log.trace("particpant: status: " + pid + "/tx/" + tid);
        return  TxUtil.response(
            TxUtil.getBaseURI() + getWork(HttpResponseCodes.SC_NOT_FOUND, pid, TxUtil.txURI + "/" + tid).getStatus());
    }

    /**
     * Performing a POST on <P-URL>/prepare causes the participant to prepare
     * the (implicitly) associated transaction. It returns a URL if
     * successful, which indicates the outcome. That URL can be probed (via
     * GET see getStatus()) and will simply return the same (implicit) information:
     *
     * <P-URL>/prepare-ok
     * <P-URL>/prepare-readonly
     * <P-URL>/prepare-notok
     */
    @POST
    @Path("{pId}/tx/{tid}/prepare/")
    public Response prepare(@PathParam("pId")String pid,
                               @PathParam("tid") @DefaultValue("")String tid)
    {
        log.debug("particpant: prepare: " + pid + "/tx/" + tid + "/prepare");
        return TxUtil.postResponse(getWork(HttpResponseCodes.SC_NOT_FOUND, pid, TxUtil.txURI + "/" + tid).prepare());
    }
    @GET
    @Path("{pId}/tx/{tid}/prepare/")
    public Response getPrepare(@PathParam("pId")String pid,
                               @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_NOT_FOUND);
    }
    @PUT
    @Path("{pId}/tx/{tid}/prepare/")
    public Response putPrepare(@PathParam("pId")String pid,
                               @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_UNAUTHORIZED);
    }

    /**
     * Performing a POST on <P-URL>/commit will cause the participant to commit
     * on behalf of the transaction. It will return a URL if successful, which
     * indicates the outcome. That URL can be probed (via GET) and will simply
     * return the same (implicit) information:
     *
     * <P-URL>/committed
     * <P-URL>/rolledback
     * <P-URL>/heuristic
     */
    @POST
    @Path("{pId}/tx/{tid}/commit/")
    public Response commit(@PathParam("pId")String pid,
                           @PathParam("tid") @DefaultValue("")String tid)
    {
        log.debug("particpant: commit: " + pid + "/tx/" + tid + "/commit");

        if (tid == null)
            throw new TransactionStatusException("no transaction");

	// get the work done on the resource in this transaction
        Work w = getWork(HttpResponseCodes.SC_NOT_FOUND, pid, TxUtil.txURI + "/" + tid);

	// and update the original resource with the changes
        return TxUtil.postResponse(w.commit(getWork(0, pid, "")));
    }
    @GET
    @Path("{pId}/tx/{tid}/commit/")
    public Response getCommit(@PathParam("pId")String pid,
                              @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_NOT_FOUND);
    }
    @PUT
    @Path("{pId}/tx/{tid}/commit/")
    public Response putCommit(@PathParam("pId")String pid,
                              @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_UNAUTHORIZED);
    }

    /**
     * Performing a POST on <P-URL>/commit will cause the participant to commit
     * on behalf of the transaction. It will return a URL if successful, which
     * indicates the outcome. That URL can be probed (via GET) and will simply
     * return the same (implicit) information:
     *
     * <P-URL>/committed
     * <P-URL>/rolledback
     * <P-URL>/heuristic<blah>
     *
     */
    @POST
    @Path("{pId}/tx/{tid}/rollback/")
    public Response rollback(@PathParam("pId")String pid,
                             @PathParam("tid") @DefaultValue("")String tid)
    {
        log.debug("particpant: rollback: " + pid + "/tx/" + tid + "/rollback");
        return TxUtil.postResponse(getWork(HttpResponseCodes.SC_NOT_FOUND, pid, TxUtil.txURI + "/" + tid).rollback());
    }
    @GET
    @Path("{pId}/tx/{tid}/rollback/")
    public Response getRollback(@PathParam("pId")String pid,
                                @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_NOT_FOUND);
    }
    @PUT
    @Path("{pId}/tx/{tid}/rollback/")
    public Response putRollback(@PathParam("pId")String pid,
                                @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_UNAUTHORIZED);
    }

    /**
     * Performing a POST on <P-URL>/forget will cause the participant to forget
     * any heuristic decision it made on behalf of the transaction.
     */
    @POST
    @Path("{pId}/tx/{tid}/forget/")
    public Response forget(@PathParam("pId")String pid,
                           @PathParam("tid") @DefaultValue("")String tid)
    {
        log.debug("POST: particpant: forget: " + pid + "/tx/" + tid + "/forget");
        return TxUtil.postResponse(getWork(HttpResponseCodes.SC_NOT_FOUND, pid, TxUtil.txURI + "/" + tid).forget());
    }

    /**
     * Performing a GET on <P-URL>/forget will return an indication of the
     * heuristic if it still exists, or 404 if it does not.
     */
    @GET
    @Path("{pId}/tx/{tid}/forget/")
    public Response getForget(@PathParam("pId")String pid,
                              @PathParam("tid") @DefaultValue("")String tid)
    {
        log.debug("GET: particpant: forget: " + pid + "/tx/" + tid + "/forget");
        Work w = getWork(HttpResponseCodes.SC_NOT_FOUND, pid, TxUtil.txURI + "/" + tid);

        if (w == null || w.getStatus() != Work.HEURISTIC)
                return TxUtil.response(HttpResponseCodes.SC_NOT_FOUND); 

        return getStatus(pid, tid, null);
    }
    @PUT
    @Path("{pId}/tx/{tid}/forget/")
    public Response putForget(@PathParam("pId")String pid,
                              @PathParam("tid") @DefaultValue("")String tid)
    {
        return TxUtil.response(HttpResponseCodes.SC_UNAUTHORIZED ); 
    }

    /*
     * Now define some resources specific to the example - namely urls
     * for retrieving and modifying the resource
     * 
     * These methods are for simulating real resources to test that the 
     * protocol is implemented correctly
     */

    @POST
    @Path("work/{pId}")
    public String modifyResource(@PathParam("pId")String pid,
                                 @QueryParam("context") @DefaultValue("")String ctx,
                                 @QueryParam("name") String name,
                                 @QueryParam("value") String value,
                                 @QueryParam("fault") String fault)
    {
        log.trace("particpant: modify " + pid + " ctx: " + ctx + " fault: " + fault);
        getWork(0, pid, ctx).put(name, value);
        
        if (fault != null)
            getWork(0, pid, ctx).put("fault", fault);

        return value;
    }

    @GET
    @Path("work/{pId}")
    public Response getResource(@PathParam("pId")String pid,
                                @QueryParam("name") String name,
                                @QueryParam("context") @DefaultValue("") String ctx)
    {
        log.trace("particpant: get: work/" + pid + " - ctx: " + ctx);
        return TxUtil.response(getWork(0, pid, ctx).get(name));
    }

    void removeParticipant(Work w)
    {
        work.remove(w.getId());
    }

    /**
     * Tell the recovery sub-system that this participant has moved
     */ 
    void moveParticipant(Work w, String newUrl)
    {
        log.trace("particpant: moving participant " + w.pid + " to " + newUrl);
        Work newWork = new Work(w);

        work.remove(w.getId());
        newWork.pid = "p99";
        String url = makeParticipantUrl(newWork);

        // remove context from old work
        w.ctx = null;

        String id = newWork.ctx == null ? newWork.pid : newWork.pid + newWork.ctx;
        work.put(id, newWork);

        try
        {
            TxUtil.doPut(new int[] {HttpResponseCodes.SC_OK}, TxUtil.rcURI + "/" + w.recCoordId, url, "URL=" + url);
        }
        catch (HttpResponseException e)
        {
            e.printStackTrace();
        }
    }

    private String makeParticipantUrl(Work w)
    {
        // use the same server as the coordinator TODO make this configurable       
        return TxUtil.getBaseURI() + new StringBuilder("participant/").append(w.pid).append("/tx/").append(w.ctx).toString().replaceAll("//", "/");
    }

    /**
     * Get the work done by this participant on behalf of the given transaction
     * If the work does not exist and the parameter statusCode is 0 then one is created
     * otherwise an HttpResponseException is thrown filled in with the value of statusCode.
     * The thrown exception will be mapped back to the required HTTP status code by the JAX RS
     * implementation.
     */
    private Work getWork(int statusCode, String pid, String ctx) throws HttpResponseException
    {
        assert ctx != null;
        assert pid != null;
        log.trace("getWork(" + pid + ", " + ctx + ')');
        int i = ctx.indexOf(TxUtil.txURI);
        if (i == -1 || i + TxUtil.txURI.length() == ctx.length())
            ctx = "";
        else
            ctx = ctx.substring(i + TxUtil.txURI.length());

        String id = pid + ctx;
        Work w = work.get(id);

        if (w == null)
        {
            if (statusCode != 0)
                throw new HttpResponseException(HttpResponseCodes.SC_OK, statusCode);

            w = new Work(this, pid, ctx);

            if (ctx.length() != 0)
            {
                try
                {
                    // enlist as a participant in this transaction context
                    String pUrl = makeParticipantUrl(w);
                    w.recCoordId = TxUtil.doPut(new int[] {HttpResponseCodes.SC_OK}, TxUtil.txURI + ctx, pUrl, "URL=" + pUrl);
                }
                catch (HttpResponseException e)
                {
                    throw new TMUnavailableException("Unable to contact TM: " + e.getMessage());
                }
                catch (Throwable t) {
                    log.info("don't know what this is");
                }
            }

            work.put(id, w);
        }

        return w;
    }

    public Participant()
    {
        /*
         * We want to remove completed work. We could remove it immediately but the spec says that status
         * urls can be probed (for an indeterminate period of time). So leave completed participants lying
         * around for a limited period before culling them.
         */
        timer = new Timer();

        int delay = 5000;   // delay for 5 sec.
        int period = 1000;  // repeat every sec.
        Timer timer = new Timer();

        timer.scheduleAtFixedRate(new TimerTask() {
            public void run() {
                Iterator<String> keys = work.keySet().iterator();

                while (keys.hasNext())
                {
                    String key = keys.next();
                    if (work.get(key).isComplete())
                        keys.remove();
                }
            }
        }, delay, period);
    }

    protected void finalize() throws Throwable
    {
        super.finalize();
        timer.cancel();
    }

    private Map<String, Work> work = new HashMap<String, Work>();
    private Timer timer;    
}
