/*
 * JBoss, Home of Professional Open Source
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2008
 * @author JBoss Inc.
 */
package org.jboss.jbossts.rts.example;

import org.jboss.jbossts.rts.provider.TransactionStatusException;
import org.jboss.jbossts.rts.provider.ResourceNotFoundException;

import java.util.Map;
import java.util.HashMap;

/**
 * Holder for isolating work done by participants from each other
 * Simulate work done inside a transaction. Its only purpose is to
 * show that work:
 * - can be isolated from other work;
 * - that is committed really does create new state;
 * - that is rolled back really does abort any changes;
 */
public class Work
{
    String pid;
    String ctx;
    String recCoordId;
    Participant bean;
    String status;
    Map<String, String> properties = new HashMap<String, String>();

    public Work(Participant bean, String pid, String ctx)
    {
        this.pid = pid;
        this.ctx = ctx;
        this.bean = bean;
    }

    public Work(Work w)
    {
        this.pid = w.pid;
        this.ctx = w.ctx;
        this.recCoordId = w.recCoordId;
        this.status = w.status;
        this.bean = w.bean;
        properties.putAll(w.properties);
    }

    public boolean isReadonly()
    {
        return properties.size() == 0;
    }

    public void put(String name, String value)
    {
        properties.put(name, value);
    }

    public String get(String name)
    {
        return properties.get(name);
    }

    public String getStatus()
    {
        return "participant/" + pid + "/tx" + ctx + '/' + status;
    }

    public String setStatus(String status)
    {
        this.status = status;

        return getStatus();
    }

    public String prepare()
    {
        validateRequest(true);
        
        return setStatus(isReadonly() ? PREPARE_RO : PREPARE_OK); // prepare-notok
    }

    static final String PREPARE_OK = "prepare-ok";
    static final String PREPARE_RO = "prepare-readonly";
    static final String ROLLEDBACK = "rolledback";
    static final String COMMIT_OK = "commit-ok";
    static final String HEURISTIC = "heuristic";

    /**
     * This method enables the tests to validate that any new state was
     * aborted.
     */
    public String rollback()
    {
        validateRequest(false);
        bean.removeParticipant(this);

        return setStatus(ROLLEDBACK); // heuristic
    }

    /**
     * This work can now be made persistent so update the original work
     * (identified by mergeTo) with the contents of this work.
     *
     * This method enables the tests to validate that any new state was
     * committed.
     */
    public String commit(Work mergeTo)
    {
        validateRequest(false);

        for (Map.Entry<String, String> e : mergeTo.properties.entrySet())
            mergeTo.put(e.getKey(), e.getValue());

        bean.removeParticipant(this);

        return setStatus(COMMIT_OK); // or rolledback or heuristic
    }

    public String forget()
    {
        validateRequest(false);
        bean.removeParticipant(this);

        return getStatus();
    }

    private void validateRequest(boolean prepare)
    {
        if (ctx == null)
            throw new TransactionStatusException("no transaction");

        String fault = properties.get("fault");

        if ((prepare && "prepare_fail".equals(fault)) || (!prepare && "commit_fail".equals(fault)))
            throw new ResourceNotFoundException("Simulated failure");
        else if ((prepare && "prepare_halt".equals(fault)) || (!prepare && "commit_halt".equals(fault)))
            Runtime.getRuntime().halt(1);

        if (prepare && fault != null && fault.startsWith("move_to"))
            bean.moveParticipant(this, fault);
    }

    public boolean isComplete()
    {
        return COMMIT_OK.equals(status) || ROLLEDBACK.equals(status);
    }

    public String getId()
    {
        return ctx == null ? pid : pid + ctx;
    }
}
