/*
 * JBoss, Home of Professional Open Source
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2008
 * @author JBoss Inc.
 */
package org.jboss.jbossts.rts.resource;

import com.arjuna.ats.arjuna.coordinator.AbstractRecord;
import com.arjuna.ats.arjuna.coordinator.RecordType;
import com.arjuna.ats.arjuna.coordinator.TwoPhaseOutcome;
import com.arjuna.ats.arjuna.common.Uid;
import com.arjuna.ats.arjuna.ObjectType;
import com.arjuna.ats.arjuna.state.OutputObjectState;
import com.arjuna.ats.arjuna.state.InputObjectState;
import com.arjuna.ats.arjuna.gandiva.ClassName;

import org.jboss.jbossts.rts.util.TxUtil;
import org.jboss.jbossts.rts.provider.HttpResponseException;
import org.jboss.resteasy.util.HttpResponseCodes;
import org.apache.log4j.Logger;

/**
 * Log record for driving participants through 2PC and recoverery
 */
public class RESTRecord extends AbstractRecord
{
    protected final static Logger log = Logger.getLogger(RESTRecord.class);
    private String url;
    private String coordinatorID;
    private String statusUrl = "";
    private String txId;
    private boolean prepared;

    public RESTRecord(String url, String txId)
    {
        super(new Uid());
        log.trace("RESTRecord url: " + url + " tx: " + txId);
        this.url = url;
        this.txId = txId;
        coordinatorID = get_uid().fileStringForm();
    }

    public static AbstractRecord create()
    {
        log.trace("RESTRecord.create");
        return new RESTRecord();
    }
    public RESTRecord()
    {
        super();
        log.trace("RESTRecord()");
    }

    protected RESTRecord(Uid u)
    {
        super(u, null, ObjectType.ANDPERSISTENT);
    }

    public int typeIs()
    {
        return RecordType.USER_DEF_FIRST0;
    }

    public ClassName className()
    {
        return new ClassName("RecordType.USER_DEF_FIRST0");
    }

    public Object value()
    {
        return statusUrl;
    }

    public void setValue(Object o)
    {
    }

    public int nestedAbort()
    {
        return TwoPhaseOutcome.FINISH_OK;
    }

    public int nestedCommit()
    {
        return TwoPhaseOutcome.FINISH_OK;
    }

    /*
    * Not subtransaction aware.
    */
    public int nestedPrepare()
    {
        return TwoPhaseOutcome.PREPARE_OK; // do nothing
    }

	private void check_suspend(Fault f)
	{
        if (fault.equals(f))
		{
            try
            {
            	log.info(f + ": for 10 seconds");
                Thread.sleep(10000);
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }
        }
	}

	private void check_halt(Fault f)
	{
        if (fault.equals(f))
		{
            log.info(f + ": halt VM");
            Runtime.getRuntime().halt(1);
		}
	}

    public int topLevelPrepare()
    {
        log.debug("prepare " + url);

        check_halt(Fault.prepare_halt);
        check_suspend(Fault.prepare_suspend);

        if (fault.equals(Fault.h_hazard))
            return TwoPhaseOutcome.HEURISTIC_HAZARD;

        if (url == null || txId == null)
            return TwoPhaseOutcome.PREPARE_READONLY;

        try
        {
            String txUrl = TxUtil.txURI + '/' + txId;
            this.statusUrl = TxUtil.doPost(new int[] {HttpResponseCodes.SC_CREATED}, url + "/prepare", "context=" + txUrl);

            prepared = true;
            return TwoPhaseOutcome.PREPARE_OK;
        }
        catch (HttpResponseException e)
        {
            log.info("participant " + url + " prepare error: " + e.getMessage());
            statusUrl = this.url + "/prepare-notok";

            return TwoPhaseOutcome.PREPARE_NOTOK;
        }
    }

    public int topLevelAbort()
    {
        log.debug("abort " + url);

        check_halt(Fault.abort_halt);
        check_suspend(Fault.abort_suspend);

        if (url == null || txId == null)
            return TwoPhaseOutcome.FINISH_ERROR;

        try {
            statusUrl = TxUtil.doPost(new int[] {HttpResponseCodes.SC_CREATED}, this.url + "/rollback");

            return TwoPhaseOutcome.FINISH_OK;
        } catch (HttpResponseException e) {
            if (log.isInfoEnabled())
                log.info("participant " + this.url + " rollback error: " + e.getMessage());
            this.statusUrl = this.url + "/heuristic"; // don't know what status the participant is in

            return TwoPhaseOutcome.FINISH_ERROR;
        }
    }

    public int topLevelCommit()
    {
        log.debug("commit " + url);

        if (url == null || txId == null)
            return TwoPhaseOutcome.PREPARE_READONLY;

        if (!prepared)
            return TwoPhaseOutcome.NOT_PREPARED;

        return topLevelOnePhaseCommit();
    }

    public int nestedOnePhaseCommit()
    {
        return TwoPhaseOutcome.FINISH_ERROR;
    }

    /**
     * For commit_one_phase we can do whatever we want since the transaction
     * outcome is whatever we want. Therefore, we do not need to save any
     * additional recoverable state, such as a reference to the transaction
     * coordinator, since it will not have an intentions list anyway.
     */

    public int topLevelOnePhaseCommit()
    {
        check_halt(Fault.commit_halt);
        check_suspend(Fault.commit_suspend);

        if (txId == null)
            return TwoPhaseOutcome.FINISH_ERROR;

//        String txUrl = TxUtil.TX_URI_PREFIX + '/' + txId;

        try
        {
            if (log.isDebugEnabled()) log.info("commiting " + this.url);
            if (!statusUrl.endsWith("prepare-readonly"))
                statusUrl = TxUtil.doPost(new int[] {HttpResponseCodes.SC_CREATED}, this.url + "/commit"); //, "context=" + txUrl));

            log.info("COMMIT OK at url: " + this.url);
            return TwoPhaseOutcome.FINISH_OK;
        }
        catch (HttpResponseException e)
        {
            if (e.getActualResponse() == HttpResponseCodes.SC_NOT_FOUND)
            {
                // the participant may have moved so check the coordinator url
                if (hasParticpantMoved())
                {
                    if (log.isDebugEnabled())
                        log.debug("participant has moved commit to new url " + this.url);
                    try
                    {
                        this.statusUrl = TxUtil.doPost(new int[] {HttpResponseCodes.SC_CREATED}, this.url + "/commit"); //, "context=" + txUrl));

            			log.info("COMMIT OK at new url: " + this.url);
                        return TwoPhaseOutcome.FINISH_OK;
                    }
                    catch (HttpResponseException e1)
                    {
                        e = e1;
                    }
                }
            }

            if (log.isInfoEnabled())
                log.debug("participant " + this.url + " commit error: " + e.getMessage());
            this.statusUrl = this.url + "/commit-notok";

            return TwoPhaseOutcome.FINISH_ERROR;
        }
    }

    /**
     * A participant tells the coordinator if it changes its URL.
     * To see if this has happened perform a GET on the recovery url which returns the
     * last known location of the participant.
     */
    private boolean hasParticpantMoved()
    {
        try
        {
            if (log.isDebugEnabled())
                log.trace("seeing if participant has moved: " + TxUtil.getBaseURI() + "recovery-coordinator/" + coordinatorID);

            // get the latest participant url by probing the recovery url:
            String url = TxUtil.doGet(new int[] {HttpResponseCodes.SC_OK}, TxUtil.getBaseURI() + "recovery-coordinator/" + coordinatorID);

            if (url != null && !url.equals(this.url))
            {
                // participant has moved so remember the new location
                this.url = url;

            	if (log.isDebugEnabled())
                	log.trace("\tnew url is " + url);
                return true;
            }
        }
        catch (HttpResponseException e)
        {
        }

        return false;
    }

    public boolean save_state(OutputObjectState os, int t)
    {
        try
        {
            os.packString(txId);
            os.packBoolean(prepared);
            os.packString(url);
            os.packString(coordinatorID);
            os.packString(statusUrl);

            return super.save_state(os, t);
        }
        catch (Exception e)
        {
            e.printStackTrace();

            return false;
        }
    }

    public boolean restore_state(InputObjectState os, int t)
    {
        try
        {
            txId = os.unpackString();
            prepared = os.unpackBoolean();
            url = os.unpackString();
            coordinatorID = os.unpackString();
            statusUrl = os.unpackString();
            log.info("restore_state " + url);

            return super.restore_state(os, t);
        }
        catch (Exception e)
        {
            return false;
        }
    }

    public String type()
    {
        return RESTRecord.typeName();
    }

    public static String typeName()
    {
        return "/StateManager/AbstractRecord/RESTRecord";
    }

    public boolean doSave()
    {
        return true;
    }

    public void merge(AbstractRecord a)
    {
    }

    public void alter(AbstractRecord a)
    {
    }

    public boolean shouldAdd(AbstractRecord a)
    {
        return (a.typeIs() == typeIs());
    }

    public boolean shouldAlter(AbstractRecord a)
    {
        return false;
    }

    public boolean shouldMerge(AbstractRecord a)
    {
        return false;
    }

    public boolean shouldReplace(AbstractRecord a)
    {
        return false;
    }

    enum Fault {
		abort_halt, abort_suspend, prepare_halt,
		prepare_suspend, commit_halt, commit_suspend,
		h_commit, h_rollback, h_hazard, h_mixed, none
	}
    Fault fault = Fault.none;

    public void setFault(String name)
    {
        for (Fault f : Fault.values())
        {
            if (f.name().equals(name))
            {
                log.trace("setFault: " + f + " url: " + url);

                fault = f;
                return;
            }
        }

        fault = Fault.none;
    }
}
